import usb.core

INIVATION_VENDOR_ID = 5419


def identify_inivation_camera():
    camera = usb.core.find(idVendor=5418)
    if camera is None:
        raise RuntimeError("No inivation camera connected")
    else:
        return f"inivation {camera.bus} {camera.address} dvx"


if __name__ == "__main__":
    print(identify_inivation_camera())



from typing import Tuple
import torch
import aestream

from durin.sensor import Sensor


class DVSSensor(Sensor[torch.Tensor]):
    def __init__(self, shape: Tuple[int, int], device: str, port: int):
        self.source = aestream.UDPInput(shape, device, port)

    def start_stream(self) -> torch.Tensor:
        return self.source.start_stream()

    def stop_stream(self) -> torch.Tensor:
        return self.source.stop_stream()

    def read(self) -> torch.Tensor:
        return self.source.read()



from durin import DurinUI, SetSensorPeriod, GetSystemInfo
import time
import sys


if __name__ == "__main__":

    # We start a connectioen to the robot
    # and can now read from and write to the robot via the variable "durin"
    # Notice the UI class, which differs from the (more efficient) standalone Durin interface
    with DurinUI("durin1.local") as durin:
        # Loop until the user quits
        is_running = True


        durin(GetSystemInfo())  # Ask to get IP address, MAC address and Durin ID
        gotSystemInfo = False

        while not gotSystemInfo:
            """ Wait until IP address, MAC address, and Durin ID are read"""
            try:
                (obs, dvs, cmd) = durin.read()
                ip = cmd.systemInfo.ip
                mac = cmd.systemInfo.mac
                id = cmd.systemInfo.id
                durin.set_ip_mac_id(ip,mac,id)
                gotSystemInfo = True

            except:
                pass

        
        while is_running:

            # Read a value from durin
            # - obs = Robot sensor observations
            # - dvs = Robot DVS data (if any)
            # - cmd = Robot responses to commands
            (obs, dvs, cmd) = durin.read()

            # We can now update our display with the observations
            durin.render_sensors(obs)

            # Read user input and quit, if asked
            is_running = durin.read_user_input()

            time.sleep(0.02)




import time
import socket
from durin import *

T_WAIT = 0.1

def example_moving(durin):

    # Move Durin using robcentric command
    (vel_x, vel_y, rot) = (-1, 2, 3)
    print(f"move_robcentric {vel_x} {vel_y} {rot}")
    reply = durin(Move(vel_x, vel_y, rot))
    print(reply)
    time.sleep(T_WAIT)

    # Move Durin using robcentric command
    (m1, m2, m3, m4) = (-5, 0, 1, 9)
    print(f"move_wheels {m1} {m2} {m3} {m4}")
    reply = durin(MoveWheels(m1, m2, m3, m4))
    print(reply)
    time.sleep(T_WAIT)

    # Move Durin using robcentric command
    (vel_x, vel_y, rot) = (6,5,2)
    print(f"move_robcentric {vel_x} {vel_y} {rot}")
    reply = durin(Move(vel_x, vel_y, rot))
    print(reply)
    time.sleep(T_WAIT)

    # Powering durin off
    reply = durin(PowerOff())
    time.sleep(T_WAIT)

def example_polling(durin):

    # Get sensory data from sensor 128
    print("Polling ToF[0:1]:")
    reply = durin(PollSensor(SENSORS["tof_a"]))
    print(reply)
    time.sleep(T_WAIT)

    # Get sensory data from sensor 129
    print("Polling ToF[2:3]")
    reply = durin(PollSensor(SENSORS["tof_b"]))
    print(reply)
    time.sleep(T_WAIT)

    # Get sensory data from sensor 130
    print("Polling ToF[4:5]")
    reply = durin(PollSensor(SENSORS["tof_c"]))
    print(reply)
    time.sleep(T_WAIT)

    # Get sensory data from sensor 131
    print("Polling ToF[6:7]")
    reply = durin(PollSensor(SENSORS["tof_d"]))
    print(reply)
    time.sleep(T_WAIT)

    # Get sensory data from sensor 132
    print("Polling BatteryCharge+Voltage) + IMU:")
    reply = durin(PollSensor(SENSORS["misc"]))
    print(reply)
    time.sleep(T_WAIT)

    # Get sensory data from sensor 133
    print("Polling UWB")
    reply = durin(PollSensor(SENSORS["uwb"]))
    print(reply)
    time.sleep(T_WAIT)


def example_streaming(durin, px_np=[], duration=10):  

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))

    
    ip_address = s.getsockname()[0]
    port_udp = 4300
    period = 100

    print(f"UDP on {ip_address}:{port_udp} ... {period}")

    # Activating/deactivating continuous streaming of sensory data on UDP
    reply = durin(StreamOn(ip_address, port_udp, period))
    count = 0
    t_start = time.time()
    t_current = t_start
    while (t_current-t_start) < duration:

        t_current = time.time()
        
        (obs, dvs) = durin.sense()
        if len(px_np) > 0:
            px_np[:,:]= obs.tof
        
        # idx = 0
        # print(obs.tof[:,idx*8:idx*8+8])
        count += 1


    print("Waiting some time ... ")
    time.sleep(2)
    reply = durin(StreamOff())



import matplotlib.pyplot as plt
import pandas as pd

# load data from CSV file into a pandas DataFrame
df = pd.read_csv('sensor statistics/stats.csv')
text_file = open("sensor statistics/stats.txt", "w")


headers = ["Acce x","Acce y","Acce z",
           "Gyro x","Gyro y","Gyro z",
           "Magn x","Magn y","Magn z",
           "Position x","Position y","Position z",
           "Voltage","UWB ID","UWB distande"]

for header in headers:
    plt.clf()
    print(header)

    # extract column of data from DataFrame
    if header == "Acce x":
        header = "# Acce x"
    data = df[header]
    mean = data.mean()
    std = data.std()

    # create histogram with 1000 bins
    plt.hist(data, bins=500)

    # set title and axis labels
    plt.title('Histogram of '+header)
    plt.xlabel('Value')
    plt.ylabel('Frequency')
    plt.xlim((mean-3*std,mean+3*std))

    # Write mean
    text_file.write(header+"\n")
    text_file.write("Mean: "+str(mean)+"\n")
    text_file.write("Standard deviation: "+str(std)+"\n")
    text_file.write("Standard deviation in %: "+str(std/mean)+"\n")
    text_file.write("\n")
 

    plt.savefig("sensor statistics/plots/"+header+".png")

text_file.close()
# # display the plot
# plt.show()


from durin import Durin, DurinUI, Move 
import time
import sys
import numpy as np

save = True

time_steps = 10000
N_sensor_values = 9 + 3 + 1 + 2         # imu + position + voltage + one beacon (ID + distance)
all_stats = np.zeros((time_steps, N_sensor_values), dtype = object)
headers = ["Acce x","Acce y","Acce z","Gyro x","Gyro y","Gyro z","Magn x","Magn y","Magn z","Position x","Position y","Position z","Voltage","UWB ID","UWB distande"]
header=','.join(headers)
print(header)

if __name__ == "__main__":
    i = 0

    with Durin("durin1.local") as durin:

        while i < time_steps:
            (obs, dvs, cmd) = durin.read()

            if obs.voltage == 0:
                # In the "boot" process, all values are 0 in a number of while-iterations.
                # We do not want to include the 0-values
                continue
            
            print(i, "/", time_steps)
            
            # Collect all relevant sensor data
            imu_values = obs.imu.reshape((9,))
            position = obs.position.reshape((3,))
            voltage = np.array([obs.voltage]).reshape((1,))
            uwb = np.array([obs.uwb[0]]).reshape((2,))
            sensor_values = np.concatenate([imu_values, position, voltage, uwb])

            # Add sensor values to the all_stats array
            for sensor_idx, sensor_value in enumerate(sensor_values):
                all_stats[(i), sensor_idx] = sensor_value

            durin(Move(0,0,0))  # So that Durin does not sleep
            i += 1


    
    # Caluculate statistics
    statistic_quantities = (["Acce x mean", "Acce x variance"],
                            ["Acce y mean", "Acce y variance"],
                            ["Acce z mean", "Acce z variance"],
                            ["Gyro x mean", "Gyro x variance"],
                            ["Gyro y mean", "Gyro y variance"],
                            ["Gyro z mean", "Gyro z variance"],
                            ["Magn x mean", "Magn x variance"],
                            ["Magn y mean", "Magn y variance"],
                            ["Magn z mean", "Magn z variance"],
                            ["Position x mean", "Position x variance"],
                            ["Position y mean", "Position y variance"],
                            ["Position z mean", "Position z variance"],
                            ["Voltage mean", "Voltage variance"],
                            ["UWB ID mean", "UWB ID variance"],
                            ["UWB distande mean", "UWB distande variance"])
    

    for j in range(len(statistic_quantities)):
        print(statistic_quantities[j][0], np.mean(all_stats[:,j]), statistic_quantities[j][1], np.var(all_stats[:,j]))

if save:
    # Save as a csv file
    np.savetxt('sensor statistics/stats.csv',all_stats,delimiter=',',header=','.join(headers), fmt='%d')







import multiprocessing

from inputs import get_gamepad, UnpluggedError

from durin.io.runnable import RunnableProducer


JOYSTICK_MIN = 21
JOYSTICK_MAX = 232
JOYSTICK_ABS = JOYSTICK_MAX - JOYSTICK_MIN
JOYSTICK_MID = JOYSTICK_ABS / 2


class Gamepad(RunnableProducer):
    def __init__(self, timeout: float = 0.05):
        context = multiprocessing.get_context("spawn")
        self.queue = context.Queue(maxsize=1)
        self.timeout = timeout
        self.values = multiprocessing.Array("f", 3)
        self.tau = 0.99

        super().__init__(self.queue, self.values)

    @staticmethod
    def norm(x):
        if x == 0:
            return int(x)
        else:
            return int(((x - JOYSTICK_MID - JOYSTICK_MIN) / JOYSTICK_ABS) * 1000)

    def produce(self, values):
        array = values.get_obj()

        # Register gamepad events
        try:
            events = get_gamepad()
            for event in events:
                stick = event.code
                state = event.state
                if state > 0:
                    if "ABS_X" in stick:
                        array[0] = state
                    elif "ABS_Y" in stick:
                        array[1] = state
                    elif "ABS_RX" in stick:
                        array[2] = state
            return Gamepad.norm(array[0]), Gamepad.norm(array[1]), Gamepad.norm(array[2])
        except UnpluggedError:
            pass
        return None



import cProfile
import pstats
import subprocess


if __name__ == '__main__':
    module_name = 'SDL2'
    # module_name = 'Pygame'
    module_name = 'Pygame 2'

    stats = pstats.Stats(f'{module_name}.prof')
    stats.strip_dirs()
    stats.sort_stats('cumtime')
    stats.print_stats(10)



from abc import abstractmethod
import logging
import multiprocessing
import subprocess
import socket
from typing import Optional

from durin.controller import dvs
from durin.io import decode
from durin.io.network import TCPProducer
from durin.io.runnable import Runnable, RunnableConsumer

logging.getLogger().setLevel(logging.DEBUG)


class Streamer:
    @abstractmethod
    def start_stream(self, host: str, port: int):
        pass

    @abstractmethod
    def stop_stream(self):
        pass


class AEStreamer(Streamer):
    def __init__(self) -> None:
        self.aestream = None
        # Test that aestream exists
        if subprocess.run(["which", "aestream"]).returncode > 0:
            raise RuntimeError("No aestream binary found on path")
        # Test that cameras exist
        try:
            self.camera_string = dvs.identify_inivation_camera()
        except RuntimeError as e:
            logging.warning("No DVX camera found on startup", e)

    def start_stream(self, host: str, port: int):
        if self.aestream is not None:
            self.stop_stream()

        # Get the camera string
        try:
            self.camera_string = dvs.identify_inivation_camera()
            logging.debug(f"Camera found at {self.camera_string}")
        except Exception as e:
            logging.warning("No camera found", e)
            return

        host_ip = ".".join([str(i) for i in host])
        command = f"aestream input {self.camera_string} output udp {host_ip} {port}"
        self.aestream = subprocess.run(
            command.split(" "),
            capture_output=True,  # stderr=subprocess.STDOUT, stdout=subprocess.PIPE
        )

        logging.debug(f"Sending DVS to {host}:{port} with command\n\t{command}")

    def stop_stream(self):
        try:
            if self.aestream is not None:
                self.aestream.terminate()
                self.aestream.wait(1)
                self.aestream.kill()
                self.aestream.wait()
            subprocess.run(["pkill", "aestream"])  # Kill remaining aestream processes
            self.aestream = None
        except Exception as e:
            logging.warning("Error when closing aestream", e)


class DVSServer(Runnable):
    def __init__(self, port: int, streamer: Optional[Streamer] = None) -> None:
        super().__init__(port)
        context = multiprocessing.get_context("spawn")
        self.buffer = context.Queue(10)
        self.streamer = streamer if streamer is not None else AEStreamer()
        self.consumer = DVSCommandConsumer(self.buffer, self.streamer)
        self.clients = []
        self.is_streaming = True

    def run(self, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.bind(("0.0.0.0", port))
            sock.listen(1)
            connection, _ = sock.accept()
            self.close_clients()

            producer = TCPProducer(self.buffer, connection)
            producer.start()
            producer.thread.join()
        except TimeoutError:
            pass
        finally:
            sock.close()

    def start(self):
        super().start()
        self.consumer.start()

    def stop(self):
        super().stop()
        self.close_clients()
        self.consumer.stop()

    def close_clients(self):
        logging.debug(f"Closing {len(self.clients)} old clients")
        for thread in self.clients:
            thread.close()
        self.clients = []


class DVSCommandConsumer(RunnableConsumer):
    def consume(self, bs, streamer: AEStreamer):
        if bs is None or len(bs) == 0:
            return
        d = decode(bs)
        if d.which == "enableStreaming":
            destination = d.enableStreaming.destination.udpOnly
            ip = destination.ip
            port = destination.port
            streamer.start_stream(ip, port)
        elif d.which == "disableStreaming":
            streamer.stop_stream()
        else:
            logging.warn("Unknown command received", d.which)


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)
    server = DVSServer(2300)
    try:
        server.start()
        server.thread.join()
    finally:
        server.stop()



from multiprocessing import Process
import multiprocessing
from time import sleep
from durin.controller import server
from durin.io import network
from durin.io.command import StreamOff, StreamOn


class MockStreamer(server.Streamer):
    def __init__(self, queue: multiprocessing.Queue) -> None:
        self.queue = queue

    def start_stream(self, host: str, port: int):
        self.queue.put(f"{host}:{port}")

    def stop_stream(self):
        self.queue.put("stop")


def test_handshake():
    q = multiprocessing.get_context("spawn").Queue(10)
    serv = server.DVSServer(3000, streamer=MockStreamer(q))
    serv.start()
    sleep(0.2)
    b = network.TCPLink("0.0.0.0", 3000)
    b.start()
    sleep(0.2)
    b.send(StreamOn("0.0.0.0", 3001, 1).encode(), 10)
    sleep(0.2)
    assert q.get(timeout=1) == "[0, 0, 0, 0]:3001"

    b.send(StreamOff().encode(), 10)
    sleep(0.2)
    assert q.get() == "stop"
    serv.stop()
    b.stop()



from setuptools import find_packages, setup

with open("requirements.txt", "r") as fp:
    requirements = fp.read().split("\n")

with open("README.md", "r") as fp:
    long_description = fp.read()

setup(
    name="durin",
    version="0.0.75",
    install_requires=requirements,
    packages=find_packages(),
    license="LGPLv3",
    maintainer="Jens E. Pedersen",
    maintainer_email="jeped@kth.se",
    extras_require={"aestream": ["aestream"]},
    scripts=["bin/durin"],
    description="Python control interface for the Durin robot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True
)


import logging
from pathlib import Path
from typing import List

import capnp

PORT_TCP = 1337
PORT_UDP = 2300

SENSORS = {
    "tof_a": 128,
    "tof_b": 129,
    "tof_c": 130,
    "tof_d": 131,
    "misc": 132,
    "uwb": 133,
}

schema = capnp.load(str((Path(__file__).parent / "schema.capnp").absolute()))

def decode(buffer) -> List[schema.DurinBase]:
    # TODO: Use packed version later
    try:
        msg = schema.DurinBase.from_bytes_packed(buffer)
        return msg
    except ValueError as e:
        logging.warn("Error in decoded message", e)
    except capnp.lib.capnp.KjException as e:
        logging.warn("Failed to deserialize", e)
        return None



import pygame
import math
import os
import random
import sys
import ctypes
import time

import numpy as np
from durin.actuator import Move

from durin.durin import Durin
from durin.io.gamepad import Gamepad
from durin import SetSensorPeriod, GetSystemInfo, EnableTofStatus


# Constants
surface_width = 200
surface_height = 200
sleep_interval = 0.02

SENSOR_PLACEMENTS = [
    (0.25, 0.01),
    (0.02, 0.03),
    (0.02, 0.4),
    (0.02, 0.7),
    (0.25, 0.78),
    (0.41, 0.7),
    (0.47, 0.4),
    (0.41, 0.02),
]

SENSOR_ROTATIONS = [0 - 90, 45 - 90, 90 - 90, 135 - 90, 180 - 90, 225 - 90, 270 - 90, 315 - 90]

# A distance (in % of screen size) constant related to the layout.
d= 0.02
x=0.68
y0 = 0.1

TITLE_PLACEMENT = (x, 0.05)

INSTR_PLACEMENT = (x, 0.05+2*d)          # Keyboard instruction (static text)

IP_PLACEMENT = (x, 0.1 + 3* d)

BATTERY_PLACEMENT = (x,0.1 + 8*d)

IMU_PLACEMENT = (x, 0.1 +11*d)  # Upper left corner

IMU_INTEG_PLACEMENT = (x, 0.1 + 17*d)

POSITION_PLACEMENT = (x,0.1+21*d)

MV_CMD_PLACEMENT = (x+7*d,0.1+21*d)         # Movement command placement

UWB_PLACEMENT = (x, 0.1+29*d)              # Upper left corner

TOF_STATUS_PLACEMENT = (x, 0.1+25*d)


class DurinUI(Durin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gamepad = Gamepad()

        self.ip = None
        self.mac = None
        self.id = None

        self.vertical = 0
        self.horizontal = 0
        self.tau = 0.9999
        self.rot = 0

    def __enter__(self):
        self.a = 0 # Just for debugging. Delete soon!

        self.set_frequency()
        self(EnableTofStatus(True))

        pygame.init()
        self.clock = pygame.time.Clock()
        self.font = pygame.font.SysFont(None, 22)
        self.big_font = pygame.font.SysFont(None, 34)


        # Set up the display
        # Get screen size
        info = pygame.display.Info()
        self.screen_width, self.screen_height = info.current_w, info.current_h-50

        self.screen = pygame.display.set_mode((0, 0), pygame.RESIZABLE)

        # Buffer screen:
        #self.back_buffer = pygame.Surface((self.screen_width, self.screen_height))

        # Make it fullscreen
        if sys.platform == "win32":
            HWND = pygame.display.get_wm_info()['window']
            SW_MAXIMIZE = 3
            ctypes.windll.user32.ShowWindow(HWND, SW_MAXIMIZE)


        # Durin Image
        resource_file = "durin/durin_birdseye.jpg"
        resource_path = os.path.join(os.getcwd(), resource_file)
        self.image = pygame.image.load(resource_path)
        self.image = pygame.transform.scale(self.image, (1.75*self.screen_width//3, self.screen_height))


        self.image_surface = pygame.Surface(self.image.get_size())
        self.image_surface.blit(self.image, (0,0))

        # Create surfaces for ToF data.
        self.surfaces = []
        for o in range(8):
            surface = pygame.Surface((surface_width, surface_height), pygame.SRCALPHA)
            self.surfaces.append(surface)

        pygame.display.update()




        return super().__enter__()

    def __exit__(self, e, b, t):
        pygame.quit()
        self.gamepad.stop()
        return super().__exit__(e, b, t)

    def read_user_input(self, allow_movement: bool = True, sleep_interval: float=0.02):
        keys = pygame.key.get_pressed()

        # Gamepad
        if not self.gamepad.queue.empty():
            x, y, r = self.gamepad.queue.get()
            self.horizontal = x
            self.vertical = y
            self.rot = r

        for event in pygame.event.get():
            if event.type == pygame.QUIT or (event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE):
                return False

            # Keyboard
            elif event.type == pygame.KEYDOWN:
                # Key pressed
                if event.key == pygame.K_UP or event.key == pygame.K_w:
                    self.vertical = 500
                elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                    self.vertical = -500
                elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
                    self.horizontal = -500
                elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
                    self.horizontal = 500
                elif event.key == pygame.K_q:
                    self.rot = 500
                elif event.key == pygame.K_e:
                    self.rot = -500

            elif event.type == pygame.KEYUP:
                # Key released
                if event.key == pygame.K_UP or event.key == pygame.K_w or event.key == pygame.K_DOWN or event.key == pygame.K_s:
                    self.vertical = 0
                if event.key == pygame.K_LEFT or event.key == pygame.K_d or event.key == pygame.K_RIGHT or event.key == pygame.K_a:
                    self.horizontal = 0
                if event.key == pygame.K_e or event.key == pygame.K_q:
                    self.rot = 0

        if allow_movement:
            self(Move(self.horizontal, self.vertical, self.rot))

        # time.sleep(sleep_interval) # Sleep to avoid sending too many commands

        return True


    def render_sensors(self, obs, size: int = 180):

        self.screen.fill((0,0,0))   # Fill screen with black
        self.screen.blit(self.image_surface, (0,0))


        # Update ToF-sensors ######################
        tofs = (np.tanh((obs.tof / 1000)) * 255).astype(np.int32)

        # Rotated surfaces
        rotated_surfaces = []

        for o in range(len(self.surfaces)):
            surface = self.surfaces[o]

            square_size = surface_width//8
            for i in range(8):
                for j in range(8):

                    left = i * square_size
                    top = j * square_size
                    square_rect = pygame.Rect(left, top, square_size, square_size)

                    color_value = tofs[o][i][j]
                    color = (color_value, color_value, color_value)
                    pygame.draw.rect(surface, color, square_rect)


            rotation_angle = SENSOR_ROTATIONS[o]
            rotated_surface = pygame.transform.rotate(surface, rotation_angle)
            rotated_surfaces.append(rotated_surface)


        for i in range(8):
            self.screen.blit(rotated_surfaces[i], (SENSOR_PLACEMENTS[i][0]*self.screen_width,SENSOR_PLACEMENTS[i][1]*self.screen_height))

        # Update UWB ######################

        uwb = obs.uwb
        #self.render_text("Becon ID\t\t\tDistance (mm)", UWB_PLACEMENT[0])


        for i in range(10):
            if uwb[i][0] != 0:
                self.render_text(str(uwb[i][0]), (UWB_PLACEMENT[0], UWB_PLACEMENT[1] + (i+1)*d))
                self.render_text(str(uwb[i][1]), (UWB_PLACEMENT[0]+ 7*d, UWB_PLACEMENT[1]+(i+1)*d))
            else:
                break



        # Update IMU ######################
        imu = obs.imu
        #type = ["Acce", "Gyro", "Magn."]
        for type in range(3):
            for xyz in range(3):
                self.render_text(str(imu[type][xyz]), (IMU_PLACEMENT[0]+(xyz+1)*3*d, IMU_PLACEMENT[1]+(type+1)*d))

        # Update ToF status ###################
        summed_status = [0] * 8
        for i in range(8):
            tot = 0
            for v in obs.tof_status[i].flat:
                if v != 0 and v != 3:
                    tot += 1
            summed_status[i] = tot

        self.render_text(f"ToF: reported faulty pixels", (TOF_STATUS_PLACEMENT[0], TOF_STATUS_PLACEMENT[1]), "o")
        self.render_text(f"0: {summed_status[0]}", (TOF_STATUS_PLACEMENT[0] + 0 * d * 2, TOF_STATUS_PLACEMENT[1] + d))
        self.render_text(f"1: {summed_status[1]}", (TOF_STATUS_PLACEMENT[0] + 1 * d * 2, TOF_STATUS_PLACEMENT[1] + d))
        self.render_text(f"2: {summed_status[2]}", (TOF_STATUS_PLACEMENT[0] + 2 * d * 2, TOF_STATUS_PLACEMENT[1] + d))
        self.render_text(f"3: {summed_status[3]}", (TOF_STATUS_PLACEMENT[0] + 3 * d * 2, TOF_STATUS_PLACEMENT[1] + d))
        self.render_text(f"4: {summed_status[4]}", (TOF_STATUS_PLACEMENT[0] + 0 * d * 2, TOF_STATUS_PLACEMENT[1] + d + d))
        self.render_text(f"5: {summed_status[5]}", (TOF_STATUS_PLACEMENT[0] + 1 * d * 2, TOF_STATUS_PLACEMENT[1] + d + d))
        self.render_text(f"6: {summed_status[6]}", (TOF_STATUS_PLACEMENT[0] + 2 * d * 2, TOF_STATUS_PLACEMENT[1] + d + d))
        self.render_text(f"7: {summed_status[7]}", (TOF_STATUS_PLACEMENT[0] + 3 * d * 2, TOF_STATUS_PLACEMENT[1] + d + d))

        # Update battery level and voltage ######################
        voltage = obs.voltage
        charge = obs.charge
        self.render_text(str(charge) + " %", BATTERY_PLACEMENT)
        self.render_text(str(voltage) + " mV", (BATTERY_PLACEMENT[0]+5*d,BATTERY_PLACEMENT[1]))


        # Update Durin position ######################
        for m in range(3):
            self.render_text(str(obs.position[m]), (POSITION_PLACEMENT[0]+2*m*d, POSITION_PLACEMENT[1]+2*d))

        # Update movement commands ################
        self.render_text(str(self.horizontal),(MV_CMD_PLACEMENT[0],MV_CMD_PLACEMENT[1]+2*d))
        self.render_text(str(self.vertical),(MV_CMD_PLACEMENT[0]+2*d,MV_CMD_PLACEMENT[1]+2*d))
        self.render_text(str(self.rot),(MV_CMD_PLACEMENT[0]+4*d,MV_CMD_PLACEMENT[1]+2*d))

        self.render_static_texts()

        # Just for debugging.
        self.a += 1
        #self.render_text("Time step (for debugging): " + str(self.a),(UWB_PLACEMENT[0],UWB_PLACEMENT[1]+10*d))

        # Update screen
        pygame.display.update()

        # self.clock.tick(25)


    def render_text(self, input_text, position, color="w", size = "small"):
        if color == "w":
            c = (255,255,255)
        elif color == "o":
            c = (255,183,91)
        elif color == "b":
            c = (100,100,255)
        elif color == "t":
            c = (255,143,161)

        if size == "small":
            text =  self.font.render(input_text, True, c)
        elif size == "big":
            text=  self.big_font.render(input_text, True, c)
        self.screen.blit(text, (position[0]*self.screen_width,position[1]*self.screen_height))

    def render_static_texts(self):
        # Static texts¨

        # Dashboard title
        self.render_text("Durin Dashboard", TITLE_PLACEMENT, "t", "big")

        # Titles for Durin IP, MAC and ID
        self.render_text("IP address", IP_PLACEMENT, "o")
        self.render_text("MAC address", (IP_PLACEMENT[0]+5*d,IP_PLACEMENT[1]), "o")
        self.render_text("Durin ID", (IP_PLACEMENT[0]+10*d,IP_PLACEMENT[1]), "o")

        # The IP, MAC and ID values
        self.render_text(str(self.ip), (IP_PLACEMENT[0],IP_PLACEMENT[1]+d))
        self.render_text(str(self.mac), (IP_PLACEMENT[0]+5*d,IP_PLACEMENT[1]+d))
        self.render_text(str(self.id), (IP_PLACEMENT[0]+10*d,IP_PLACEMENT[1]+d))

        # IMU-related titles
        self.render_text("IMU data",(IMU_PLACEMENT[0],IMU_PLACEMENT[1]), "o")
        self.render_text("x",(IMU_PLACEMENT[0]+3*d,IMU_PLACEMENT[1]), "b")
        self.render_text("y",(IMU_PLACEMENT[0]+6*d,IMU_PLACEMENT[1]),"b")
        self.render_text("z",(IMU_PLACEMENT[0]+9*d,IMU_PLACEMENT[1]),"b")
        self.render_text("Acce",(IMU_PLACEMENT[0],IMU_PLACEMENT[1]+d),"b")
        self.render_text("Gyro",(IMU_PLACEMENT[0],IMU_PLACEMENT[1]+2*d),"b")
        self.render_text("Magn",(IMU_PLACEMENT[0],IMU_PLACEMENT[1]+3*d),"b")

        # Integrated IMU title
        self.render_text("Integrated IMU data", (IMU_INTEG_PLACEMENT), "o")

        # Battery related titles
        self.render_text("Battery level",(BATTERY_PLACEMENT[0],BATTERY_PLACEMENT[1]-d), "o")
        self.render_text("Voltage",(BATTERY_PLACEMENT[0]+5*d,BATTERY_PLACEMENT[1]-d), "o")

        # Durin coordinate titles
        self.render_text("Durin coordinates",(POSITION_PLACEMENT[0],POSITION_PLACEMENT[1]), "o")
        self.render_text("x",(POSITION_PLACEMENT[0],POSITION_PLACEMENT[1]+d),"b")
        self.render_text("y",(POSITION_PLACEMENT[0]+2*d,POSITION_PLACEMENT[1]+d),"b")
        self.render_text("z",(POSITION_PLACEMENT[0]+4*d,POSITION_PLACEMENT[1]+d),"b")

        # Instructions for keyboard shortcuts
        keyboard_instruction = "Use the keys w, a, s, d or arrow keys or a gamepad"
        keyboard_instruction2 = "to move Durin. Press q or e for rotations."
        self.render_text(keyboard_instruction,INSTR_PLACEMENT)
        self.render_text(keyboard_instruction2,(INSTR_PLACEMENT[0],INSTR_PLACEMENT[1]+d))

        # Movement commands titles
        self.render_text("Movement commands", MV_CMD_PLACEMENT, "o")
        self.render_text("x",(MV_CMD_PLACEMENT[0],MV_CMD_PLACEMENT[1]+d),"b")
        self.render_text("y",(MV_CMD_PLACEMENT[0]+2*d,MV_CMD_PLACEMENT[1]+d),"b")
        self.render_text("rot",(MV_CMD_PLACEMENT[0]+4*d,MV_CMD_PLACEMENT[1]+d),"b")

        # UWB related titles
        self.render_text("UWB ID", (UWB_PLACEMENT[0],UWB_PLACEMENT[1]), "o")
        self.render_text("Distance (mm)", (UWB_PLACEMENT[0]+7*d,UWB_PLACEMENT[1]), "o")

    def set_ip_mac_id(self, ip, mac, id):
        self.ip = ip
        self.mac = mac
        self.id = id

    def set_frequency(self):
        # The sensor frequencies in Hz
        sensor_frequencies = (["Imu", 50],
                              ["Position", 50],
                              ["SystemStatus", 1],
                              ["Uwb", 50],
                              ["Tof", 50],
                              )


        for sensor in sensor_frequencies:
            self(SetSensorPeriod(sensor[0],1000/sensor[1]))    # Frequency (Hz) to period (ms)




import cProfile
import subprocess
import pstats



def profile_module(module_name):
    subprocess.call(['python', '-m', 'cProfile', '-o', f'{module_name}.prof', f'{module_name}.py'])

module_name = 'durin/examples/CPU_test'
profile_module(module_name)

stats = pstats.Stats(f'{module_name}.prof')
stats.strip_dirs()
stats.sort_stats('tottime')
stats.print_stats(10)

